#define DEVICE 257
#define CARD 258
#define ANONYMOUS 259
#define TUPLE 260
#define VERSION 261
#define FUNCTION 262
#define BIND 263
#define MODULE 264
#define OPTS 265
#define START 266
#define STOP 267
#define REGION 268
#define JEDEC 269
#define DTYPE 270
#define DEFAULT 271
#define MTD 272
#define INCLUDE 273
#define EXCLUDE 274
#define RESERVE 275
#define IRQ_NO 276
#define PORT 277
#define MEMORY 278
#define STRING 279
#define NUMBER 280
typedef union {
    char *str;
    int num;
    struct device_info_t *device;
    struct card_info_t *card;
    struct mtd_ident_t *mtd;
    struct adjust_list_t *adjust;
} YYSTYPE;
extern YYSTYPE yylval;
